
/****************************************************************/
/*  
     ***   ***  ***  *** ***     *** ***   ******  ****  ***
     ***   ***  ***  *** ***     *** **   ***  *** ***** ***
     ***   ***  ***  *** ***     *****    ******** *** *****
      *** ***   ***  *** ******* *** **   ***  *** ***  ****
       *****     *****   ******* *** ***  ***  *** ***   ***
	   
	 Vulkan Graphics API: in 20 Minutes (Coffee Break Series)
	 Beginners Taste

     Minimal working examples (MWE) to build upon and 
	 modify to gain an understanding of the API

	 Extensions to the Vulkan Graphics Coffee Break Series
	   
/*
/****************************************************************/

#pragma once

#pragma comment(linker, "/subsystem:windows")
#include <windows.h>


#include "vulkan.h"


#include "helpers.h"


#include <stdio.h>
#include <stdlib.h>

using namespace std;
#include <vector>
#include <array>


extern VkCoffeeDevice *				coffeeDevice;
extern VkCoffeeSwapChain*			coffeeSwapChain;

class CoffeeDrawQuad
{
public:

	VkDescriptorSet					descriptorSet			= 0; 
	VkDescriptorSetLayout			descriptorSetLayout		= 0;
	VkPipelineLayout				pipelineLayout			= 0;
	VkDescriptorPool				descriptorPool			= 0;
	VkCommandPool					cmdPool					= 0;
	std::vector<VkCommandBuffer>	drawCmdBuffers;
	VkPipeline						pipeline				= 0;
	VkSampler						sampler					= 0;
	VkShaderModule					shaderModuleVS			= 0;
	VkShaderModule					shaderModulePS			= 0;
	float							offset					= 0;

	struct Vertex 
	{
		float pos[3];
		float uv[2];
		float col[3];
		float normal[3];
	};

	

	struct QuadMesh
	{
		struct VerticeDescription
		{
			VkPipelineVertexInputStateCreateInfo           inputState;
			std::vector<VkVertexInputBindingDescription>   bindingDescriptions;
			std::vector<VkVertexInputAttributeDescription> attributeDescriptions;
		};
			
		VkBuffer			vbuf  = VK_NULL_HANDLE;
		VkDeviceMemory		vmem  = VK_NULL_HANDLE;
		size_t				vsize = 0;
	
		VkBuffer			ibuf  = VK_NULL_HANDLE;
		VkDeviceMemory		imem  = VK_NULL_HANDLE;
		size_t				isize = 0;

		uint32_t			indexCount;
		Vector3				dim;
		VerticeDescription  verticeDescription;
	} quadMesh;

	





	CoffeeDrawQuad(float offset)
	{
		this->offset = offset;

		SetupSampler();
		SetupDescriptorPool();
		SetupDescriptorSetLayout();
		SetupDescriptorSet();
		SetupVertexDescriptions();
	
		PreparePipelines();	
		
		CreateCommandPool();
		CreateCommandBuffers();
	}
	
	void SetupSampler()
	{
		VkSamplerCreateInfo samplerCreateInfo	= {};
		samplerCreateInfo.sType					= VK_STRUCTURE_TYPE_SAMPLER_CREATE_INFO;
		samplerCreateInfo.magFilter				= VK_FILTER_LINEAR;
		samplerCreateInfo.minFilter				= VK_FILTER_LINEAR;
		samplerCreateInfo.mipmapMode			= VK_SAMPLER_MIPMAP_MODE_LINEAR;
		samplerCreateInfo.addressModeU			= VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE;
		samplerCreateInfo.addressModeV			= VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE;
		samplerCreateInfo.addressModeW			= VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE;
		samplerCreateInfo.mipLodBias			= 0;
		samplerCreateInfo.anisotropyEnable		= VK_FALSE;
		samplerCreateInfo.minLod				= 0;
		samplerCreateInfo.maxLod				= 5;
		samplerCreateInfo.borderColor			= VK_BORDER_COLOR_FLOAT_TRANSPARENT_BLACK;
		samplerCreateInfo.unnormalizedCoordinates = VK_FALSE;

		VkResult result = 
		vkCreateSampler( coffeeDevice->device, 
						 &samplerCreateInfo, 
						 NULL, 
						 &sampler );
		DBG_ASSERT_VULKAN_MSG( result, 
			"Failed to create sampler." );
	}

	void SetupVertexDescriptions()
	{
		float s = 0.5f;
		// Setup vertices for a single uv-mapped quad
#define QUAD_COLOUR_NORMAL { 1.0f, 1.0f, 1.0f }, { 0.0f, 0.0f, 1.0f }
		std::vector<Vertex> vertexBuffer =
		{
			{ { s*1.0f+offset, s*1.0f, 0.0f },{ 1.0f, 1.0f }, QUAD_COLOUR_NORMAL },
			{ { s*0.0f+offset, s*1.0f, 0.0f },{ 0.0f, 1.0f }, QUAD_COLOUR_NORMAL },
			{ { s*0.0f+offset, s*0.0f, 0.0f },{ 0.0f, 0.0f }, QUAD_COLOUR_NORMAL },
			{ { s*1.0f+offset, s*0.0f, 0.0f },{ 1.0f, 0.0f }, QUAD_COLOUR_NORMAL }
		};
#undef QUAD_COLOUR_NORMAL

		CreateBuffer(
			coffeeDevice,
			VK_BUFFER_USAGE_VERTEX_BUFFER_BIT,
			VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT,
			vertexBuffer.size() * sizeof(Vertex),
			vertexBuffer.data(),
			&quadMesh.vbuf,
			&quadMesh.vmem);

		// Setup indices
		std::vector<uint32_t> indexBuffer = { 0,1,2, 2,3,0 };
		quadMesh.indexCount = indexBuffer.size();

		CreateBuffer(
			coffeeDevice,
			VK_BUFFER_USAGE_INDEX_BUFFER_BIT,
			VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT,
			indexBuffer.size() * sizeof(uint32_t),
			indexBuffer.data(),
			&quadMesh.ibuf,
			&quadMesh.imem);


		// Binding description
		quadMesh.verticeDescription.bindingDescriptions.resize(1);

		VkVertexInputBindingDescription vInputBindDescription = {};
		vInputBindDescription.binding   = VERTEX_BUFFER_BIND_ID;
		vInputBindDescription.stride    = sizeof(Vertex);
		vInputBindDescription.inputRate = VK_VERTEX_INPUT_RATE_VERTEX;
		quadMesh.verticeDescription.bindingDescriptions[0] = vInputBindDescription;

		// Attribute descriptions
		quadMesh.verticeDescription.attributeDescriptions.resize(4);

		// Location 0 : Position
		quadMesh.verticeDescription.attributeDescriptions[0].location = 0;
		quadMesh.verticeDescription.attributeDescriptions[0].binding  = VERTEX_BUFFER_BIND_ID;
		quadMesh.verticeDescription.attributeDescriptions[0].format   = VK_FORMAT_R32G32B32_SFLOAT;
		quadMesh.verticeDescription.attributeDescriptions[0].offset   = 0;

		// Location 1 : Texture coordinates
		quadMesh.verticeDescription.attributeDescriptions[1].location = 1;
		quadMesh.verticeDescription.attributeDescriptions[1].binding  = VERTEX_BUFFER_BIND_ID;
		quadMesh.verticeDescription.attributeDescriptions[1].format   = VK_FORMAT_R32G32_SFLOAT;
		quadMesh.verticeDescription.attributeDescriptions[1].offset   = sizeof(float) * 3;

		// Location 2 : Color
		quadMesh.verticeDescription.attributeDescriptions[2].location = 2;
		quadMesh.verticeDescription.attributeDescriptions[2].binding  = VERTEX_BUFFER_BIND_ID;
		quadMesh.verticeDescription.attributeDescriptions[2].format   = VK_FORMAT_R32G32B32_SFLOAT;
		quadMesh.verticeDescription.attributeDescriptions[2].offset   = sizeof(float) * 5;

		// Location 3 : Normal
		quadMesh.verticeDescription.attributeDescriptions[3].location = 3;
		quadMesh.verticeDescription.attributeDescriptions[3].binding  = VERTEX_BUFFER_BIND_ID;
		quadMesh.verticeDescription.attributeDescriptions[3].format   = VK_FORMAT_R32G32B32_SFLOAT;
		quadMesh.verticeDescription.attributeDescriptions[3].offset   = sizeof(float) * 8;

		VkPipelineVertexInputStateCreateInfo pipelineVertexInputStateCreateInfo = {};
		pipelineVertexInputStateCreateInfo.sType = VK_STRUCTURE_TYPE_PIPELINE_VERTEX_INPUT_STATE_CREATE_INFO;
		pipelineVertexInputStateCreateInfo.pNext = NULL;

		quadMesh.verticeDescription.inputState                                 = pipelineVertexInputStateCreateInfo;
		quadMesh.verticeDescription.inputState.vertexBindingDescriptionCount   = quadMesh.verticeDescription.bindingDescriptions.size();
		quadMesh.verticeDescription.inputState.pVertexBindingDescriptions      = quadMesh.verticeDescription.bindingDescriptions.data();
		quadMesh.verticeDescription.inputState.vertexAttributeDescriptionCount = quadMesh.verticeDescription.attributeDescriptions.size();
		quadMesh.verticeDescription.inputState.pVertexAttributeDescriptions    = quadMesh.verticeDescription.attributeDescriptions.data();
	}// End SetupVertexDescriptions(..)



	VkBool32 CreateBuffer(VkCoffeeDevice*		coffeeDevice,
	                      VkBufferUsageFlags    usageFlags, 
	                      VkMemoryPropertyFlags memoryPropertyFlags, 
						  VkDeviceSize          size, 
						  void *                data, 
						  VkBuffer *            buffer, 
						  VkDeviceMemory *      memory)
	{
		VkMemoryRequirements memReqs;
		VkMemoryAllocateInfo memAlloc = {};
		memAlloc.sType           = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
		memAlloc.pNext           = NULL;
		memAlloc.allocationSize  = 0;
		memAlloc.memoryTypeIndex = 0;

		VkBufferCreateInfo bufferCreateInfo = {};
		bufferCreateInfo.sType = VK_STRUCTURE_TYPE_BUFFER_CREATE_INFO;
		bufferCreateInfo.pNext = NULL;
		bufferCreateInfo.usage = usageFlags;
		bufferCreateInfo.size  = size;
		bufferCreateInfo.flags = 0;


		VK_CHECK_RESULT(vkCreateBuffer(coffeeDevice->device, &bufferCreateInfo, nullptr, buffer));

		vkGetBufferMemoryRequirements(coffeeDevice->device, *buffer, &memReqs);
		memAlloc.allocationSize  = memReqs.size;
		memAlloc.memoryTypeIndex = coffeeDevice->GetMemoryType(memReqs.memoryTypeBits, memoryPropertyFlags);
		VK_CHECK_RESULT(vkAllocateMemory(coffeeDevice->device, &memAlloc, nullptr, memory));
		if (data != nullptr)
		{
			void *mapped;
			VK_CHECK_RESULT(vkMapMemory(coffeeDevice->device, *memory, 0, size, 0, &mapped));
			memcpy(mapped, data, (size_t)size);
			vkUnmapMemory(coffeeDevice->device, *memory);
		}
		VK_CHECK_RESULT(vkBindBufferMemory(coffeeDevice->device, *buffer, *memory, 0));

		return true;
	}

	void SetupDescriptorPool()
	{
		VkDescriptorPoolSize descriptorPoolSize1	= {};
		descriptorPoolSize1.type					= VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
		descriptorPoolSize1.descriptorCount			= 50;

		std::vector<VkDescriptorPoolSize> poolSizes =
		{
			descriptorPoolSize1
		};

		VkDescriptorPoolCreateInfo descriptorPoolInfo = {};
		descriptorPoolInfo.sType					= VK_STRUCTURE_TYPE_DESCRIPTOR_POOL_CREATE_INFO;
		descriptorPoolInfo.pNext					= NULL;
		descriptorPoolInfo.poolSizeCount			= poolSizes.size();
		descriptorPoolInfo.pPoolSizes				= poolSizes.data();
		descriptorPoolInfo.maxSets					= 10;


		VK_CHECK_RESULT(vkCreateDescriptorPool(coffeeDevice->device, &descriptorPoolInfo, nullptr, &descriptorPool));
	}

	void SetupDescriptorSetLayout()
	{
		std::vector<VkDescriptorSetLayoutBinding> setLayoutBindings;
		VkDescriptorSetLayoutBinding setLayoutBinding1	= {};
		setLayoutBinding1.descriptorType				= VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
		setLayoutBinding1.stageFlags					= VK_SHADER_STAGE_FRAGMENT_BIT;
		setLayoutBinding1.binding						= 1; // Binding number from shader file
		setLayoutBinding1.descriptorCount				= 1; // Number of descriptors (e.g., arrays)
		setLayoutBinding1.pImmutableSamplers			= 0;

		// Binding 1 : Fragment shader image sampler
		setLayoutBindings.push_back(setLayoutBinding1);


		// Shaded layouts (only use first layout binding)
		VkDescriptorSetLayoutCreateInfo descriptorSetLayoutCreateInfo = {};
		descriptorSetLayoutCreateInfo.sType			= VK_STRUCTURE_TYPE_DESCRIPTOR_SET_LAYOUT_CREATE_INFO;
		descriptorSetLayoutCreateInfo.pNext			= NULL;
		descriptorSetLayoutCreateInfo.pBindings		= setLayoutBindings.data();
		descriptorSetLayoutCreateInfo.bindingCount	= setLayoutBindings.size();
		VK_CHECK_RESULT(vkCreateDescriptorSetLayout(coffeeDevice->device, &descriptorSetLayoutCreateInfo, nullptr, &descriptorSetLayout));

		VkPipelineLayoutCreateInfo pipelineLayoutCreateInfo = {};
		pipelineLayoutCreateInfo.sType				= VK_STRUCTURE_TYPE_PIPELINE_LAYOUT_CREATE_INFO;
		pipelineLayoutCreateInfo.pNext				= NULL;
		pipelineLayoutCreateInfo.setLayoutCount		= 1;
		pipelineLayoutCreateInfo.pSetLayouts		= &descriptorSetLayout;
		VK_CHECK_RESULT(vkCreatePipelineLayout(coffeeDevice->device, &pipelineLayoutCreateInfo, nullptr, &pipelineLayout));
	}



	void SetupDescriptorSet()
	{
		// Descriptor set
		VkDescriptorSetAllocateInfo allocInfo		= {};
		allocInfo.sType								= VK_STRUCTURE_TYPE_DESCRIPTOR_SET_ALLOCATE_INFO;
		allocInfo.pNext								= NULL;
		allocInfo.descriptorPool					= descriptorPool;
		allocInfo.pSetLayouts						= &descriptorSetLayout;
		allocInfo.descriptorSetCount				= 1;

		VK_CHECK_RESULT(vkAllocateDescriptorSets(coffeeDevice->device, &allocInfo, &descriptorSet));
	}

	
	void PreparePipelines()
	{
		VkPipelineInputAssemblyStateCreateInfo inputAssemblyState = {};
		inputAssemblyState.sType					= VK_STRUCTURE_TYPE_PIPELINE_INPUT_ASSEMBLY_STATE_CREATE_INFO;
		inputAssemblyState.topology					= VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST;
		inputAssemblyState.flags					= 0;
		inputAssemblyState.primitiveRestartEnable	= VK_FALSE;

		VkPipelineRasterizationStateCreateInfo rasterizationState = {};
		rasterizationState.sType					= VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_STATE_CREATE_INFO;
		rasterizationState.polygonMode				= VK_POLYGON_MODE_FILL;
		rasterizationState.cullMode					= VK_CULL_MODE_NONE;
		rasterizationState.frontFace				= VK_FRONT_FACE_COUNTER_CLOCKWISE;
		rasterizationState.flags					= 0;
		rasterizationState.depthClampEnable			= VK_FALSE;
		rasterizationState.lineWidth				= 1.0f;

		VkPipelineViewportStateCreateInfo viewportState = {};
		viewportState.sType							= VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_STATE_CREATE_INFO;
		viewportState.viewportCount					= 1;
		viewportState.scissorCount					= 1;
		viewportState.flags							= 0;

		std::vector<VkDynamicState> dynamicStateEnables = {
			VK_DYNAMIC_STATE_VIEWPORT,
			VK_DYNAMIC_STATE_SCISSOR
		};

		VkPipelineDynamicStateCreateInfo dynamicState = {};
		dynamicState.sType							= VK_STRUCTURE_TYPE_PIPELINE_DYNAMIC_STATE_CREATE_INFO;
		dynamicState.pDynamicStates					= dynamicStateEnables.data(),
		dynamicState.dynamicStateCount				= dynamicStateEnables.size();

		// Load shaders
		std::array<VkPipelineShaderStageCreateInfo, 2> shaderStages;
		shaderStages[0] = LoadShader("./quad.vert.spv", VK_SHADER_STAGE_VERTEX_BIT,   &shaderModuleVS);
		shaderStages[1] = LoadShader("./quad.frag.spv", VK_SHADER_STAGE_FRAGMENT_BIT, &shaderModulePS);


		VkGraphicsPipelineCreateInfo pipelineCreateInfo = {};
		pipelineCreateInfo.sType					= VK_STRUCTURE_TYPE_GRAPHICS_PIPELINE_CREATE_INFO;
		pipelineCreateInfo.pNext					= NULL;
		pipelineCreateInfo.layout					= pipelineLayout;
		pipelineCreateInfo.renderPass				= 0;
		pipelineCreateInfo.flags					= 0;
		pipelineCreateInfo.pVertexInputState		= &quadMesh.verticeDescription.inputState;
		pipelineCreateInfo.pInputAssemblyState		= &inputAssemblyState;
		pipelineCreateInfo.pRasterizationState		= &rasterizationState;
		pipelineCreateInfo.pColorBlendState			= 0;
		pipelineCreateInfo.pMultisampleState		= 0;
		pipelineCreateInfo.pViewportState			= &viewportState;
		pipelineCreateInfo.pDepthStencilState		= NULL;		// Don't need a depth buffer
		pipelineCreateInfo.pDynamicState			= &dynamicState;
		pipelineCreateInfo.stageCount				= shaderStages.size();
		pipelineCreateInfo.pStages					= shaderStages.data();


		VK_CHECK_RESULT(vkCreateGraphicsPipelines(coffeeDevice->device, VK_NULL_HANDLE, 1, &pipelineCreateInfo, nullptr, &pipeline));
	}


	void CreateCommandPool()
	{
		VkCommandPoolCreateInfo cmdPoolInfo			= {};
		cmdPoolInfo.sType							= VK_STRUCTURE_TYPE_COMMAND_POOL_CREATE_INFO;
		cmdPoolInfo.queueFamilyIndex				= coffeeSwapChain->queueNodeIndex;
		cmdPoolInfo.flags							= VK_COMMAND_POOL_CREATE_RESET_COMMAND_BUFFER_BIT;
		VK_CHECK_RESULT(vkCreateCommandPool(coffeeDevice->device, &cmdPoolInfo, nullptr, &cmdPool));
	}

	void CreateCommandBuffers()
	{
		// Double buffering the command buffers (one is being drawn, the other is
		// getting updated)
		drawCmdBuffers.resize( 2 );

		VkCommandBufferAllocateInfo cmdBufAllocateInfo = {};
		cmdBufAllocateInfo.sType					= VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO;
		cmdBufAllocateInfo.commandPool				= cmdPool;
		cmdBufAllocateInfo.level					= VK_COMMAND_BUFFER_LEVEL_PRIMARY;
		cmdBufAllocateInfo.commandBufferCount		= static_cast<uint32_t>(drawCmdBuffers.size());

		VK_CHECK_RESULT(vkAllocateCommandBuffers(coffeeDevice->device, &cmdBufAllocateInfo, drawCmdBuffers.data()));
	}

	void BuildCommandBuffers(int screenWidth, 
							 int screenHeight)
	{
		VkCommandBufferBeginInfo cmdBufInfo			= {};
		cmdBufInfo.sType							= VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO;
		cmdBufInfo.pNext							= NULL;

		for (int32_t i = 0; i < (int32_t)drawCmdBuffers.size(); ++i)
		{
			VK_CHECK_RESULT(vkBeginCommandBuffer(drawCmdBuffers[i], &cmdBufInfo));

			VkViewport viewport						= {};
			viewport.width							= (float)screenWidth;
			viewport.height							= (float)screenHeight;
			viewport.minDepth						= 0;
			viewport.maxDepth						= 1;
			vkCmdSetViewport(drawCmdBuffers[i], 0, 1, &viewport);

			VkRect2D scissor      = {};
			scissor.extent.width  = screenWidth;
			scissor.extent.height = screenHeight;
			scissor.offset.x      = 0;
			scissor.offset.y      = 0;
			vkCmdSetScissor(drawCmdBuffers[i], 0, 1, &scissor);

			VkDeviceSize offsets[1] = { 0 };

			vkCmdBindDescriptorSets(drawCmdBuffers[i], VK_PIPELINE_BIND_POINT_GRAPHICS, pipelineLayout, 0, 1, &descriptorSet, 0, NULL);
			vkCmdBindPipeline(drawCmdBuffers[i], VK_PIPELINE_BIND_POINT_GRAPHICS, pipeline);
			vkCmdBindVertexBuffers(drawCmdBuffers[i], VERTEX_BUFFER_BIND_ID, 1, &quadMesh.vbuf, offsets);
			vkCmdBindIndexBuffer(drawCmdBuffers[i], quadMesh.ibuf, 0, VK_INDEX_TYPE_UINT32);
			vkCmdDrawIndexed(drawCmdBuffers[i], quadMesh.indexCount, 1, 0, 0, 0);

			VK_CHECK_RESULT(vkEndCommandBuffer(drawCmdBuffers[i]));
		}
	}

	void Render(int			frameBufferIndex, 
	            int			screenWidth,
				int         screenHeight,
	            VkImageView imageView)
	{
		VkDescriptorImageInfo descriptorImageInfo;
		descriptorImageInfo.imageLayout = VK_IMAGE_LAYOUT_GENERAL;
		descriptorImageInfo.imageView   = imageView;
		descriptorImageInfo.sampler     = sampler;

		#if 1
		{
			VkWriteDescriptorSet wd[1];
			memset(wd, 0, sizeof(wd));

			wd[0].sType            = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
			wd[0].dstSet           = descriptorSet;
			wd[0].dstBinding       = 1;
			wd[0].dstArrayElement  = 0; 
			wd[0].descriptorCount  = 1;
			wd[0].descriptorType   = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
			wd[0].pImageInfo       = &descriptorImageInfo;
			wd[0].pBufferInfo      = NULL;
			wd[0].pTexelBufferView = NULL;

			vkUpdateDescriptorSets( coffeeDevice->device, 1, wd, 0, NULL );
		}
		#endif

		// Warning - if you don't rebuild the command buffer - then changes to the
		// descriptorset won't take effect - for example, if you update the texture/image
		// information (via vkUpdateDescriptorSets) - you won't see any changes if you 
		// don't rebuild the command buffer
		BuildCommandBuffers(screenWidth, screenHeight);

		VkFence           renderFence     = {};
		VkFenceCreateInfo fenceCreateInfo = {};
		fenceCreateInfo.sType			  = VK_STRUCTURE_TYPE_FENCE_CREATE_INFO;
		vkCreateFence( coffeeDevice->device, &fenceCreateInfo, NULL, &renderFence );

		VkSubmitInfo si			= {};
		si.sType                = VK_STRUCTURE_TYPE_SUBMIT_INFO;
		si.waitSemaphoreCount   = 0;
		si.pWaitSemaphores      = VK_NULL_HANDLE;
		si.pWaitDstStageMask    = NULL;
		si.commandBufferCount   = 1;
		si.pCommandBuffers      = &drawCmdBuffers[frameBufferIndex];
		si.signalSemaphoreCount = 0;
		si.pSignalSemaphores    = VK_NULL_HANDLE;

		vkQueueSubmit( coffeeDevice->queue, 1, &si, renderFence );

		vkWaitForFences( coffeeDevice->device, 1, &renderFence, VK_TRUE, UINT64_MAX );
		vkDestroyFence( coffeeDevice->device, renderFence, NULL );
	}


};










