/****************************************************************/
/*  
     ***   ***  ***  *** ***     *** ***   ******  ****  ***
     ***   ***  ***  *** ***     *** **   ***  *** ***** ***
     ***   ***  ***  *** ***     *****    ******** *** *****
      *** ***   ***  *** ******* *** **   ***  *** ***  ****
       *****     *****   ******* *** ***  ***  *** ***   ***
	   
	 Vulkan Graphics API: in 20 Minutes (Coffee Break Series)
	 Beginners Taste

     Minimalistic examples to build upon and experiment with 
	 to develop an understanding of the underlying API

	   
/*
/****************************************************************/
/*
     About:
	 Minimal Vulkan demo (i.e., stripped down to the basics).
	 Remember to compile and include the shader
	 files (vertex and pixel shaders).
	 e.g., run the compilershaders.bat

	 Extension Chapters 0X - Debug Textured Quad

	 Display textured quad on the screen (overlapping) with
	 texture information - show loaded/generated textures

*/




#pragma comment(lib, "C:\\VulkanSDK\\1.0.17.0\\Bin32\\vulkan-1.lib")


#define _CRT_SECURE_NO_WARNINGS // fopen

#pragma comment(linker, "/subsystem:windows")
#include <windows.h>

#include "vulkan.h"

#include <stdio.h>
#include <stdlib.h>

using namespace std;
#include <vector>
#include <array>


#define VERTEX_BUFFER_BIND_ID     0

#include "helpers.h"
#include "tgatexture.h"
#include "drawquad.h"



/**************************************************************/
/*
  Main class that does all the work (VulkanDemo)


*/
/**************************************************************/




VkCoffeeDevice *				 coffeeDevice				= NULL;
VkCoffeeSwapChain*				 coffeeSwapChain			= NULL;
VkCoffeeTexture*				 coffeeTexture				= NULL;
CoffeeDrawQuad*					 coffeeDrawQuad				= NULL;
VkFormat						 colorformat				= VK_FORMAT_B8G8R8A8_UNORM;
VkFormat						 depthFormat				= VK_FORMAT_D32_SFLOAT;
std::vector<VkCoffeeTGATexture*> coffeeTGATextures;



class VulkanDemo
{
public:

	VkPipeline						 pipeline				= 0;
	VkPipelineLayout				 pipelineLayout			= 0;
	VkDescriptorSet					 descriptorSet			= 0;
	VkDescriptorSetLayout			 descriptorSetLayout	= 0;
	VkDescriptorPool				 descriptorPool			= VK_NULL_HANDLE;

	VkPipelineStageFlags			 submitPipelineStages	= VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT;
	VkCommandPool					 cmdPool				= 0;
	std::vector<VkCommandBuffer>	 clearCmdBuffers;
	std::vector<VkCommandBuffer>	 drawCmdBuffers;
	std::vector<VkFramebuffer>		 frameBuffers;
	uint32_t						 currentBuffer			= 0;
	VkRenderPass					 renderPass				= 0;
	VkShaderModule					 shaderModuleVS			= 0;
	VkShaderModule					 shaderModulePS			= 0;


	VkClearColorValue				 defaultClearColor		= { { 0.025f, 0.025f, 0.225f, 1.0f } };

	// Use to adjust mouse rotation and zoom speed
	float							 rotationSpeed			= 0.01f;
	float							 zoomSpeed				= 10.0f;
	float							 zoom					= -6;


	Vector3							 rotation				= Vector3(0,  0,    0);
	Vector3							 cameraPos				= Vector3(0,  1,    0);
	Vector3							 mousePos				= Vector3(0,  0,    0);
	Vector3							 meshPos				= Vector3(0, -1.5f, 0);
	Vector3							 meshRot				= Vector3(0,  0,    0);
									 
	const char*						 title					= "Rendering a quad and texture";
	const char*						 name					= "Vulkan Graphics API: in 20 Minutes (Coffee Break Series)";
	HWND							 windowHWND				= 0;
	HINSTANCE						 windowInstance			= 0;
	uint32_t						 width					= 1280;
	uint32_t						 height					= 720;

	struct 
	{
		VkImage        image;
		VkDeviceMemory mem;
		VkImageView    view;
	} depthStencil;

	struct UniformData 
	{
		VkBuffer               buffer;
		VkDeviceMemory         memory;
		VkDescriptorBufferInfo descriptor;
		uint32_t               allocSize;
		void*                  mapped     = nullptr;
	};

	UniformData uniformData;

	struct UBO 
	{
		Matrix4  model;
		Matrix4  view;
		Matrix4  projection;
	} ubo;

	struct 
	{
		VkCoffeeMesh*	cube;
		VkCoffeeMesh*	quad;
		VkCoffeeMesh*	plane;
	} meshes;





	// Todo - release resources/assets/memory
	void ReleaseVulkan()
	{
	}
	


	void CreateCommandBuffers()
	{
		// Create one command buffer for each swap chain image and reuse for rendering
		{
		DBG_ASSERT(clearCmdBuffers.size()==0);
		clearCmdBuffers.resize(coffeeSwapChain->imageCount);

		VkCommandBufferAllocateInfo cmdBufAllocateInfo = {};
		cmdBufAllocateInfo.sType              = VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO;
		cmdBufAllocateInfo.commandPool        = cmdPool;
		cmdBufAllocateInfo.level              = VK_COMMAND_BUFFER_LEVEL_PRIMARY;
		cmdBufAllocateInfo.commandBufferCount = static_cast<uint32_t>(clearCmdBuffers.size());

		VK_CHECK_RESULT(vkAllocateCommandBuffers(coffeeDevice->device, &cmdBufAllocateInfo, clearCmdBuffers.data()));
		}

		{
		drawCmdBuffers.resize(coffeeSwapChain->imageCount);

		VkCommandBufferAllocateInfo cmdBufAllocateInfo = {};
		cmdBufAllocateInfo.sType              = VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO;
		cmdBufAllocateInfo.commandPool        = cmdPool;
		cmdBufAllocateInfo.level              = VK_COMMAND_BUFFER_LEVEL_PRIMARY;
		cmdBufAllocateInfo.commandBufferCount = static_cast<uint32_t>(drawCmdBuffers.size());

		VK_CHECK_RESULT(vkAllocateCommandBuffers(coffeeDevice->device, &cmdBufAllocateInfo, drawCmdBuffers.data()));
		}

	}








	void SetupDescriptorPool()
	{
		VkDescriptorPoolSize descriptorPoolSize0	= {};
		descriptorPoolSize0.type					= VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER;
		descriptorPoolSize0.descriptorCount			= 6;

		VkDescriptorPoolSize descriptorPoolSize1	= {};
		descriptorPoolSize1.type					= VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
		descriptorPoolSize1.descriptorCount			= 5;

		std::vector<VkDescriptorPoolSize> poolSizes =
		{
			descriptorPoolSize0,
			descriptorPoolSize1
		};

		VkDescriptorPoolCreateInfo descriptorPoolInfo = {};
		descriptorPoolInfo.sType					= VK_STRUCTURE_TYPE_DESCRIPTOR_POOL_CREATE_INFO;
		descriptorPoolInfo.pNext					= NULL;
		descriptorPoolInfo.poolSizeCount			= poolSizes.size();
		descriptorPoolInfo.pPoolSizes				= poolSizes.data();
		descriptorPoolInfo.maxSets					= 5;


		VK_CHECK_RESULT(vkCreateDescriptorPool(coffeeDevice->device, &descriptorPoolInfo, nullptr, &descriptorPool));
	}

	VkDescriptorSetLayoutCreateInfo DescriptorSetLayoutCreateInfo(
					const VkDescriptorSetLayoutBinding* pBindings,
					uint32_t bindingCount)
	{
		VkDescriptorSetLayoutCreateInfo descriptorSetLayoutCreateInfo = {};
		descriptorSetLayoutCreateInfo.sType			= VK_STRUCTURE_TYPE_DESCRIPTOR_SET_LAYOUT_CREATE_INFO;
		descriptorSetLayoutCreateInfo.pNext			= NULL;
		descriptorSetLayoutCreateInfo.pBindings		= pBindings;
		descriptorSetLayoutCreateInfo.bindingCount	= bindingCount;
		return descriptorSetLayoutCreateInfo;
	}

	VkPipelineLayoutCreateInfo PipelineLayoutCreateInfo(
					const VkDescriptorSetLayout* pSetLayouts,
					uint32_t setLayoutCount)
	{
		VkPipelineLayoutCreateInfo pipelineLayoutCreateInfo = {};
		pipelineLayoutCreateInfo.sType				= VK_STRUCTURE_TYPE_PIPELINE_LAYOUT_CREATE_INFO;
		pipelineLayoutCreateInfo.pNext				= NULL;
		pipelineLayoutCreateInfo.setLayoutCount		= setLayoutCount;
		pipelineLayoutCreateInfo.pSetLayouts		= pSetLayouts;
		return pipelineLayoutCreateInfo;
	}

	void SetupDescriptorSetLayout()
	{
		std::vector<VkDescriptorSetLayoutBinding> setLayoutBindings;
		VkDescriptorSetLayoutCreateInfo descriptorLayoutInfo;
		VkPipelineLayoutCreateInfo pipelineLayoutInfo;

		VkDescriptorSetLayoutBinding setLayoutBinding0 = {};
		setLayoutBinding0.descriptorType			= VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER;
		setLayoutBinding0.stageFlags				= VK_SHADER_STAGE_VERTEX_BIT;
		setLayoutBinding0.binding					= 0;
		setLayoutBinding0.descriptorCount			= 1; 

		VkDescriptorSetLayoutBinding setLayoutBinding1 = {};
		setLayoutBinding1.descriptorType			= VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
		setLayoutBinding1.stageFlags				= VK_SHADER_STAGE_FRAGMENT_BIT;
		setLayoutBinding1.binding					= 1;
		setLayoutBinding1.descriptorCount			= 1;


		// Binding 0 : Vertex shader uniform buffer
		setLayoutBindings.push_back(setLayoutBinding0);
		// Binding 1 : Fragment shader image sampler
		setLayoutBindings.push_back(setLayoutBinding1);


		// Shaded layouts (only use first layout binding)
		descriptorLayoutInfo = DescriptorSetLayoutCreateInfo(setLayoutBindings.data(), setLayoutBindings.size());
		VK_CHECK_RESULT(vkCreateDescriptorSetLayout(coffeeDevice->device, &descriptorLayoutInfo, nullptr, &descriptorSetLayout));

		pipelineLayoutInfo = PipelineLayoutCreateInfo(&descriptorSetLayout, 1);
		VK_CHECK_RESULT(vkCreatePipelineLayout(coffeeDevice->device, &pipelineLayoutInfo, nullptr, &pipelineLayout));
	}



	void SetupDescriptorSet()
	{
		// Descriptor set
		VkDescriptorSetAllocateInfo allocInfo		= {};
		allocInfo.sType								= VK_STRUCTURE_TYPE_DESCRIPTOR_SET_ALLOCATE_INFO;
		allocInfo.pNext								= NULL;
		allocInfo.descriptorPool					= descriptorPool;
		allocInfo.pSetLayouts						= &descriptorSetLayout;
		allocInfo.descriptorSetCount				= 1;

		VK_CHECK_RESULT(vkAllocateDescriptorSets(coffeeDevice->device, &allocInfo, &descriptorSet));

		VkWriteDescriptorSet wd[2];
		memset(wd, 0, sizeof(wd));
		
		wd[0].sType            = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
		wd[0].dstSet           = descriptorSet;
		wd[0].dstBinding       = 0;
		wd[0].dstArrayElement  = 0;
		wd[0].descriptorCount  = 1;
		wd[0].descriptorType   = VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER;
		wd[0].pImageInfo       = NULL;
		wd[0].pBufferInfo      = &uniformData.descriptor;
		wd[0].pTexelBufferView = NULL;


		wd[1].sType            = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
		wd[1].dstSet           = descriptorSet;
		wd[1].dstBinding       = 1;
		wd[1].dstArrayElement  = 0; 
		wd[1].descriptorCount  = 1;
		wd[1].descriptorType   = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
		wd[1].pImageInfo       = &coffeeTexture->descriptorImageInfo;
		wd[1].pBufferInfo      = NULL;
		wd[1].pTexelBufferView = NULL;

		vkUpdateDescriptorSets( coffeeDevice->device, 2, wd, 0, NULL );
	}


	void PreparePipelines()
	{
		VkPipelineInputAssemblyStateCreateInfo inputAssemblyState = {};
		inputAssemblyState.sType                  = VK_STRUCTURE_TYPE_PIPELINE_INPUT_ASSEMBLY_STATE_CREATE_INFO;
		inputAssemblyState.topology               = VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST;
		inputAssemblyState.flags                  = 0;
		inputAssemblyState.primitiveRestartEnable = VK_FALSE;

		VkPipelineRasterizationStateCreateInfo rasterizationState = {};
		rasterizationState.sType				= VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_STATE_CREATE_INFO;
		rasterizationState.polygonMode			= VK_POLYGON_MODE_FILL;
		rasterizationState.cullMode				= VK_CULL_MODE_NONE;//VK_CULL_MODE_BACK_BIT;
		rasterizationState.frontFace			= VK_FRONT_FACE_COUNTER_CLOCKWISE;
		rasterizationState.flags				= 0;
		rasterizationState.depthClampEnable		= VK_FALSE;
		rasterizationState.lineWidth			= 1.0f;

		VkPipelineColorBlendAttachmentState blendAttachmentState = {};
		blendAttachmentState.colorWriteMask		= 0xf;
		blendAttachmentState.blendEnable		= VK_FALSE;

		VkPipelineColorBlendStateCreateInfo colorBlendState = {};
		colorBlendState.sType					= VK_STRUCTURE_TYPE_PIPELINE_COLOR_BLEND_STATE_CREATE_INFO;
		colorBlendState.pNext					= NULL;
		colorBlendState.attachmentCount			= 1;
		colorBlendState.pAttachments			= &blendAttachmentState;

		VkPipelineDepthStencilStateCreateInfo depthStencilState = {};
		depthStencilState.sType					= VK_STRUCTURE_TYPE_PIPELINE_DEPTH_STENCIL_STATE_CREATE_INFO;
		depthStencilState.depthTestEnable		= VK_TRUE;
		depthStencilState.depthWriteEnable		= VK_TRUE;
		depthStencilState.depthCompareOp		= VK_COMPARE_OP_LESS_OR_EQUAL;
		depthStencilState.front					= depthStencilState.back;
		depthStencilState.back.compareOp		= VK_COMPARE_OP_ALWAYS;

		VkPipelineViewportStateCreateInfo viewportState = {};
		viewportState.sType						= VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_STATE_CREATE_INFO;
		viewportState.viewportCount				= 1;
		viewportState.scissorCount				= 1;
		viewportState.flags						= 0;

		VkPipelineMultisampleStateCreateInfo multisampleState = {};
		multisampleState.sType					= VK_STRUCTURE_TYPE_PIPELINE_MULTISAMPLE_STATE_CREATE_INFO;
		multisampleState.rasterizationSamples	= VK_SAMPLE_COUNT_1_BIT;

		std::vector<VkDynamicState> dynamicStateEnables = {
			VK_DYNAMIC_STATE_VIEWPORT,
			VK_DYNAMIC_STATE_SCISSOR
		};

		VkPipelineDynamicStateCreateInfo dynamicState = {};
		dynamicState.sType						= VK_STRUCTURE_TYPE_PIPELINE_DYNAMIC_STATE_CREATE_INFO;
		dynamicState.pDynamicStates				= dynamicStateEnables.data(),
		dynamicState.dynamicStateCount			= dynamicStateEnables.size();

		// Load shaders
		std::array<VkPipelineShaderStageCreateInfo, 2> shaderStages;

		VkGraphicsPipelineCreateInfo pipelineCreateInfo = {};
		pipelineCreateInfo.sType				= VK_STRUCTURE_TYPE_GRAPHICS_PIPELINE_CREATE_INFO;
		pipelineCreateInfo.pNext				= NULL;
		pipelineCreateInfo.layout				= pipelineLayout;
		pipelineCreateInfo.renderPass			= renderPass;
		pipelineCreateInfo.flags				= 0;

		pipelineCreateInfo.pVertexInputState	= &meshes.cube->verticeDescription.inputState;
		pipelineCreateInfo.pInputAssemblyState	= &inputAssemblyState;
		pipelineCreateInfo.pRasterizationState	= &rasterizationState;
		pipelineCreateInfo.pColorBlendState		= &colorBlendState;
		pipelineCreateInfo.pMultisampleState	= &multisampleState;
		pipelineCreateInfo.pViewportState		= &viewportState;
		pipelineCreateInfo.pDepthStencilState	= &depthStencilState;
		pipelineCreateInfo.pDynamicState		= &dynamicState;
		pipelineCreateInfo.stageCount			= shaderStages.size();
		pipelineCreateInfo.pStages				= shaderStages.data();
		pipelineCreateInfo.layout				= pipelineLayout;

		// Scene
		shaderStages[0] = LoadShader("./model.vert.spv", VK_SHADER_STAGE_VERTEX_BIT,   &shaderModuleVS);
		shaderStages[1] = LoadShader("./model.frag.spv", VK_SHADER_STAGE_FRAGMENT_BIT, &shaderModulePS);

		VK_CHECK_RESULT(vkCreateGraphicsPipelines(coffeeDevice->device, VK_NULL_HANDLE, 1, &pipelineCreateInfo, nullptr, &pipeline));
	}

	// Prepare and initialize uniform buffer containing shader uniforms
	void PrepareUniformBuffers()
	{
		// Mesh vertex shader uniform buffer block
		CreateBuffer(
			VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT,
			VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT,
			sizeof(ubo),
			nullptr,
			&uniformData.buffer,
			&uniformData.memory,
			&uniformData.descriptor);
	}




	void CreateCommandPool()
	{
		VkCommandPoolCreateInfo cmdPoolInfo					= {};
		cmdPoolInfo.sType									= VK_STRUCTURE_TYPE_COMMAND_POOL_CREATE_INFO;
		cmdPoolInfo.queueFamilyIndex						= coffeeSwapChain->queueNodeIndex;
		cmdPoolInfo.flags									= VK_COMMAND_POOL_CREATE_RESET_COMMAND_BUFFER_BIT;
		VK_CHECK_RESULT(vkCreateCommandPool(coffeeDevice->device, &cmdPoolInfo, nullptr, &cmdPool));
	}


	void SetupRenderPass()
	{
		// Setup depth stencil pass
		VkImageCreateInfo image								= {};
		image.sType											= VK_STRUCTURE_TYPE_IMAGE_CREATE_INFO;
		image.pNext											= NULL;
		image.imageType										= VK_IMAGE_TYPE_2D;
		image.format										= depthFormat;
		image.extent										= { width, height, 1 };
		image.mipLevels										= 1;
		image.arrayLayers									= 1;
		image.samples										= VK_SAMPLE_COUNT_1_BIT;
		image.tiling										= VK_IMAGE_TILING_OPTIMAL;
		image.usage											= VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT | VK_IMAGE_USAGE_TRANSFER_SRC_BIT;
		image.flags											= 0;

		VkMemoryAllocateInfo mem_alloc						= {};
		mem_alloc.sType										= VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
		mem_alloc.pNext										= NULL;
		mem_alloc.allocationSize							= 0;
		mem_alloc.memoryTypeIndex							= 0;

		VkImageViewCreateInfo depthStencilView				= {};
		depthStencilView.sType								= VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO;
		depthStencilView.pNext								= NULL;
		depthStencilView.viewType							= VK_IMAGE_VIEW_TYPE_2D;
		depthStencilView.format								= depthFormat;
		depthStencilView.flags								= 0;
		depthStencilView.subresourceRange					= {};
		depthStencilView.subresourceRange.aspectMask		= VK_IMAGE_ASPECT_DEPTH_BIT | VK_IMAGE_ASPECT_STENCIL_BIT;
		depthStencilView.subresourceRange.baseMipLevel		= 0;
		depthStencilView.subresourceRange.levelCount		= 1;
		depthStencilView.subresourceRange.baseArrayLayer	= 0;
		depthStencilView.subresourceRange.layerCount		= 1;

		VkMemoryRequirements memReqs;

		VK_CHECK_RESULT(vkCreateImage(coffeeDevice->device, &image, nullptr, &depthStencil.image));
		vkGetImageMemoryRequirements(coffeeDevice->device, depthStencil.image, &memReqs);
		mem_alloc.allocationSize			= memReqs.size;
		mem_alloc.memoryTypeIndex			= coffeeDevice->GetMemoryType(memReqs.memoryTypeBits, VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT);
		VK_CHECK_RESULT(vkAllocateMemory(coffeeDevice->device, &mem_alloc, nullptr, &depthStencil.mem));
		VK_CHECK_RESULT(vkBindImageMemory(coffeeDevice->device, depthStencil.image, depthStencil.mem, 0));

		depthStencilView.image = depthStencil.image;
		VK_CHECK_RESULT(vkCreateImageView(coffeeDevice->device, &depthStencilView, nullptr, &depthStencil.view));


		std::array<VkAttachmentDescription, 2> attachments = {};
		// Color attachment
		attachments[0].format						= colorformat;
		attachments[0].samples						= VK_SAMPLE_COUNT_1_BIT;
		attachments[0].loadOp						= VK_ATTACHMENT_LOAD_OP_CLEAR;
		attachments[0].storeOp						= VK_ATTACHMENT_STORE_OP_STORE;
		attachments[0].stencilLoadOp				= VK_ATTACHMENT_LOAD_OP_DONT_CARE;
		attachments[0].stencilStoreOp				= VK_ATTACHMENT_STORE_OP_DONT_CARE;
		attachments[0].initialLayout				= VK_IMAGE_LAYOUT_UNDEFINED;
		attachments[0].finalLayout					= VK_IMAGE_LAYOUT_PRESENT_SRC_KHR;
		// Depth attachment
		attachments[1].format						= depthFormat;
		attachments[1].samples						= VK_SAMPLE_COUNT_1_BIT;
		attachments[1].loadOp						= VK_ATTACHMENT_LOAD_OP_CLEAR;
		attachments[1].storeOp						= VK_ATTACHMENT_STORE_OP_STORE;
		attachments[1].stencilLoadOp				= VK_ATTACHMENT_LOAD_OP_DONT_CARE;
		attachments[1].stencilStoreOp				= VK_ATTACHMENT_STORE_OP_DONT_CARE;
		attachments[1].initialLayout				= VK_IMAGE_LAYOUT_UNDEFINED;
		attachments[1].finalLayout					= VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL;

		VkAttachmentReference colorReference		= {};
		colorReference.attachment					= 0;
		colorReference.layout						= VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL;

		VkAttachmentReference depthReference		= {};
		depthReference.attachment					= 1;
		depthReference.layout						= VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL;

		VkSubpassDescription subpassDescription    = {};
		subpassDescription.pipelineBindPoint       = VK_PIPELINE_BIND_POINT_GRAPHICS;
		subpassDescription.colorAttachmentCount    = 1;
		subpassDescription.pColorAttachments       = &colorReference;
		subpassDescription.pDepthStencilAttachment = &depthReference;
		subpassDescription.inputAttachmentCount    = 0;
		subpassDescription.pInputAttachments       = nullptr;
		subpassDescription.preserveAttachmentCount = 0;
		subpassDescription.pPreserveAttachments    = nullptr;
		subpassDescription.pResolveAttachments     = nullptr;

		// Subpass dependencies for layout transitions
		std::array<VkSubpassDependency, 2> dependencies;

		dependencies[0].srcSubpass                 = VK_SUBPASS_EXTERNAL;
		dependencies[0].dstSubpass                 = 0;
		dependencies[0].srcStageMask               = VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT;
		dependencies[0].dstStageMask               = VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT;
		dependencies[0].srcAccessMask              = VK_ACCESS_MEMORY_READ_BIT;
		dependencies[0].dstAccessMask              = VK_ACCESS_COLOR_ATTACHMENT_READ_BIT | VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT;
		dependencies[0].dependencyFlags            = VK_DEPENDENCY_BY_REGION_BIT;

		dependencies[1].srcSubpass                 = 0;
		dependencies[1].dstSubpass                 = VK_SUBPASS_EXTERNAL;
		dependencies[1].srcStageMask               = VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT;
		dependencies[1].dstStageMask               = VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT;
		dependencies[1].srcAccessMask              = VK_ACCESS_COLOR_ATTACHMENT_READ_BIT | VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT;
		dependencies[1].dstAccessMask              = VK_ACCESS_MEMORY_READ_BIT;
		dependencies[1].dependencyFlags            = VK_DEPENDENCY_BY_REGION_BIT;

		VkRenderPassCreateInfo renderPassInfo = {};
		renderPassInfo.sType                       = VK_STRUCTURE_TYPE_RENDER_PASS_CREATE_INFO;
		renderPassInfo.attachmentCount             = static_cast<uint32_t>(attachments.size());
		renderPassInfo.pAttachments                = attachments.data();
		renderPassInfo.subpassCount                = 1;
		renderPassInfo.pSubpasses                  = &subpassDescription;
		renderPassInfo.dependencyCount             = static_cast<uint32_t>(dependencies.size());
		renderPassInfo.pDependencies               = dependencies.data();

		VK_CHECK_RESULT(vkCreateRenderPass(coffeeDevice->device, &renderPassInfo, nullptr, &renderPass));
	}

	

	
	void SetupFrameBuffer()
	{
		VkImageView attachments[2];

		// Depth/Stencil attachment is the same for all frame buffers
		attachments[1] = depthStencil.view;

		VkFramebufferCreateInfo frameBufferCreateInfo = {};
		frameBufferCreateInfo.sType					= VK_STRUCTURE_TYPE_FRAMEBUFFER_CREATE_INFO;
		frameBufferCreateInfo.pNext					= NULL;
		frameBufferCreateInfo.renderPass			= renderPass;
		frameBufferCreateInfo.attachmentCount		= 2;
		frameBufferCreateInfo.pAttachments			= attachments;
		frameBufferCreateInfo.width					= width;
		frameBufferCreateInfo.height				= height;
		frameBufferCreateInfo.layers				= 1;

		// Create frame buffers for every swap chain image
		frameBuffers.resize(coffeeSwapChain->imageCount);
		for (uint32_t i = 0; i < frameBuffers.size(); i++)
		{
			attachments[0] = coffeeSwapChain->imageViews[i];
			VK_CHECK_RESULT(vkCreateFramebuffer(coffeeDevice->device, &frameBufferCreateInfo, nullptr, &frameBuffers[i]));
		}
	}


	void SetupVulkan()
	{
		// Vulkan device creation
		DBG_ASSERT(coffeeDevice==NULL);
		coffeeDevice = new VkCoffeeDevice();

		DBG_ASSERT(coffeeSwapChain==NULL);
		coffeeSwapChain = new VkCoffeeSwapChain(coffeeDevice->instance, 
												coffeeDevice->device,
												coffeeDevice->physicalDevice,
												windowHWND);
		width  = coffeeSwapChain->width;
		height = coffeeSwapChain->height;
		
		// Simple texture class
		DBG_ASSERT(coffeeTexture==NULL);
		coffeeTexture = new VkCoffeeTexture(coffeeDevice->device,
								            coffeeDevice->physicalDevice);

		coffeeTGATextures.push_back( new VkCoffeeTGATexture( "./assets/check.tga", 
															 coffeeDevice->device,
															 coffeeDevice->physicalDevice) );

		coffeeTGATextures.push_back( new VkCoffeeTGATexture( "./assets/test.tga", 
															 coffeeDevice->device,
															 coffeeDevice->physicalDevice) );

		meshes.cube  = new VkCoffeeMesh( coffeeDevice, VkCoffeeMesh::Cube  );
		meshes.quad  = new VkCoffeeMesh( coffeeDevice, VkCoffeeMesh::Quad  );
		meshes.plane = new VkCoffeeMesh( coffeeDevice, VkCoffeeMesh::Plane );

		coffeeDrawQuad  = new CoffeeDrawQuad(0.0f);

		CreateCommandPool();
		CreateCommandBuffers();
		/*
			call `BuildCommandBuffer' to update the commands we want to perform
			such as drawing and vertices
		*/

		SetupRenderPass();
		SetupFrameBuffer();
		

		PrepareUniformBuffers();
		/*
			call 'UpdateUniformBuffers' to update the data in the uniform buffer
			in the render loop
		*/

		SetupDescriptorPool();
		SetupDescriptorSetLayout();
		SetupDescriptorSet();

		PreparePipelines();
	}


	

	void BuildCommandBuffers()
	{
		VkCommandBufferBeginInfo cmdBufInfo			= {};
		cmdBufInfo.sType							= VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO;
		cmdBufInfo.pNext							= NULL;

		VkClearValue clearValues[2];
		clearValues[0].color						= defaultClearColor;
		clearValues[1].depthStencil					= { 1.0f, 0 };

		VkRenderPassBeginInfo renderPassBeginInfo	= {};
		renderPassBeginInfo.sType                   = VK_STRUCTURE_TYPE_RENDER_PASS_BEGIN_INFO;
		renderPassBeginInfo.pNext                   = NULL;
		renderPassBeginInfo.renderPass              = renderPass;
		renderPassBeginInfo.renderArea.offset.x     = 0;
		renderPassBeginInfo.renderArea.offset.y     = 0;
		renderPassBeginInfo.renderArea.extent.width = width;
		renderPassBeginInfo.renderArea.extent.height= height;
		renderPassBeginInfo.clearValueCount         = 2;
		renderPassBeginInfo.pClearValues            = clearValues;

		for (int32_t i = 0; i < 2; ++i)
		{
			// Set target frame buffer
			renderPassBeginInfo.framebuffer = frameBuffers[i]; 

			VK_CHECK_RESULT(vkBeginCommandBuffer(clearCmdBuffers[i], &cmdBufInfo));

			vkCmdBeginRenderPass(clearCmdBuffers[i], &renderPassBeginInfo, VK_SUBPASS_CONTENTS_INLINE);

			VkViewport viewport		= {};
			viewport.width			= (float)width;
			viewport.height			= (float)height;
			viewport.minDepth		= 0;
			viewport.maxDepth		= 1;
			vkCmdSetViewport(clearCmdBuffers[i], 0, 1, &viewport);

			VkRect2D scissor		= {};
			scissor.extent.width	= width;
			scissor.extent.height	= height;
			scissor.offset.x		= 0;
			scissor.offset.y		= 0;
			vkCmdSetScissor(clearCmdBuffers[i], 0, 1, &scissor);

			vkCmdEndRenderPass(clearCmdBuffers[i]);

			VK_CHECK_RESULT(vkEndCommandBuffer(clearCmdBuffers[i]));
		}


		#if 1
		for (int32_t i = 0; i < 2; ++i)
		{
			renderPassBeginInfo.framebuffer = frameBuffers[i]; 

			VK_CHECK_RESULT(vkBeginCommandBuffer(drawCmdBuffers[i], &cmdBufInfo));

			// Notice - we don't apply the `vkCmdBeginRenderPass' cmd as we do
			// with the clearCmdBuffer

			VkViewport viewport = {};
			viewport.width    = (float)width;
			viewport.height   = (float)height;
			viewport.minDepth = 0;
			viewport.maxDepth = 1;
			vkCmdSetViewport(drawCmdBuffers[i], 0, 1, &viewport);

			VkRect2D scissor      = {};
			scissor.extent.width  = width;
			scissor.extent.height = height;
			scissor.offset.x      = 0;
			scissor.offset.y      = 0;
			vkCmdSetScissor(drawCmdBuffers[i], 0, 1, &scissor);

			VkDeviceSize offsets[1] = { 0 };

			// Plane
			vkCmdBindDescriptorSets(drawCmdBuffers[i], VK_PIPELINE_BIND_POINT_GRAPHICS, pipelineLayout, 0, 1, &descriptorSet, 0, NULL);
			vkCmdBindPipeline(drawCmdBuffers[i], VK_PIPELINE_BIND_POINT_GRAPHICS, pipeline);

			vkCmdBindVertexBuffers(drawCmdBuffers[i], VERTEX_BUFFER_BIND_ID, 1, &meshes.plane->vbuf, offsets);
			vkCmdBindIndexBuffer(drawCmdBuffers[i], meshes.plane->ibuf, 0, VK_INDEX_TYPE_UINT32);
			vkCmdDrawIndexed(drawCmdBuffers[i], meshes.plane->indexCount, 1, 0, 0, 0);

			// Model
			vkCmdBindDescriptorSets(drawCmdBuffers[i], VK_PIPELINE_BIND_POINT_GRAPHICS, pipelineLayout, 0, 1, &descriptorSet, 0, NULL);
			vkCmdBindPipeline(drawCmdBuffers[i], VK_PIPELINE_BIND_POINT_GRAPHICS, pipeline);

			vkCmdBindVertexBuffers(drawCmdBuffers[i], VERTEX_BUFFER_BIND_ID, 1, &meshes.cube->vbuf, offsets);
			vkCmdBindIndexBuffer(drawCmdBuffers[i], meshes.cube->ibuf, 0, VK_INDEX_TYPE_UINT32);
			vkCmdDrawIndexed(drawCmdBuffers[i], meshes.cube->indexCount, 1, 0, 0, 0);

			VK_CHECK_RESULT(vkEndCommandBuffer(drawCmdBuffers[i]));
		}
		#endif
	}

	void Render()
	{
		if (!coffeeSwapChain)
			return;

		
		VK_CHECK_RESULT(coffeeSwapChain->AcquireNextImage(coffeeDevice->device, 
		                                                  NULL, 
												          &currentBuffer));

		// Debug timer to show increments
		static int   imageIndx = 0;
		static float delay     = 0;
		delay += 0.002f;

		char buf[256];
		sprintf(buf, "%s - %.2f, delay", name, delay);
		::SetWindowText(windowHWND, buf); 

		if ( delay > 1 )
		{
			delay = 0;
			imageIndx++;
			imageIndx = imageIndx % coffeeTGATextures.size();
		}
		VkDescriptorImageInfo* descriptorImageInfo = &coffeeTGATextures[imageIndx]->descriptorImageInfo;


		// Swap textures on the fly
		if ( 1 ) 
		{
			VkWriteDescriptorSet wd[2];
			memset(wd, 0, sizeof(wd));
		
			wd[0].sType            = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
			wd[0].dstSet           = descriptorSet;
			wd[0].dstBinding       = 0;
			wd[0].dstArrayElement  = 0;
			wd[0].descriptorCount  = 1;
			wd[0].descriptorType   = VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER;
			wd[0].pImageInfo       = NULL;
			wd[0].pBufferInfo      = &uniformData.descriptor;
			wd[0].pTexelBufferView = NULL;


			wd[1].sType            = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
			wd[1].dstSet           = descriptorSet;
			wd[1].dstBinding       = 1;
			wd[1].dstArrayElement  = 0; 
			wd[1].descriptorCount  = 1;
			wd[1].descriptorType   = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
			wd[1].pImageInfo       = descriptorImageInfo;
			#if 0 // Debug
			wd[1].pImageInfo       = &coffeeTexture->descriptorImageInfo;
			#endif
			wd[1].pBufferInfo      = NULL;
			wd[1].pTexelBufferView = NULL;

			vkUpdateDescriptorSets( coffeeDevice->device, 2, wd, 0, NULL );
		}
		

		UpdateUniformBuffers();

		// **1** Configure what we want to draw and connect the textures with
		// the shader (and lighting)
		BuildCommandBuffers();

		// **2** Rendering to the screen
		// Clear screen
		{
		VkFence           renderFence;
		VkFenceCreateInfo fenceCreateInfo = {};
		fenceCreateInfo.sType = VK_STRUCTURE_TYPE_FENCE_CREATE_INFO;
		vkCreateFence( coffeeDevice->device, &fenceCreateInfo, NULL, &renderFence );

		VkSubmitInfo si = {};
		si.sType                = VK_STRUCTURE_TYPE_SUBMIT_INFO;
		si.waitSemaphoreCount   = 0;
		si.pWaitSemaphores      = VK_NULL_HANDLE;
		si.pWaitDstStageMask    = NULL;
		si.commandBufferCount   = 1;
		si.pCommandBuffers      = &clearCmdBuffers[currentBuffer];
		si.signalSemaphoreCount = 0;
		si.pSignalSemaphores    = VK_NULL_HANDLE;

		vkQueueSubmit( coffeeDevice->queue, 1, &si, renderFence );

		vkWaitForFences( coffeeDevice->device, 1, &renderFence, VK_TRUE, UINT64_MAX );
		vkDestroyFence( coffeeDevice->device, renderFence, NULL );
		}

#if 1
		// Draw simple scene
		{
		meshPos.x = 0;
		meshPos.z = 0;
		UpdateUniformBuffers();

		VkFence           renderFence;
		VkFenceCreateInfo fenceCreateInfo = {};
		fenceCreateInfo.sType = VK_STRUCTURE_TYPE_FENCE_CREATE_INFO;
		vkCreateFence( coffeeDevice->device, &fenceCreateInfo, NULL, &renderFence );

		VkSubmitInfo si = {};
		si.sType                = VK_STRUCTURE_TYPE_SUBMIT_INFO;
		si.waitSemaphoreCount   = 0;
		si.pWaitSemaphores      = VK_NULL_HANDLE;
		si.pWaitDstStageMask    = NULL;
		si.commandBufferCount   = 1;
		si.pCommandBuffers      = &drawCmdBuffers[currentBuffer];
		si.signalSemaphoreCount = 0;
		si.pSignalSemaphores    = VK_NULL_HANDLE;

		vkQueueSubmit( coffeeDevice->queue, 1, &si, renderFence );

		vkWaitForFences( coffeeDevice->device, 1, &renderFence, VK_TRUE, UINT64_MAX );
		vkDestroyFence( coffeeDevice->device, renderFence, NULL );
		}
#endif
				
		// Render 2D debug quad on screen (with texture on it)
		{
			coffeeDrawQuad->Render(currentBuffer, 
								   width,
								   height,
								   descriptorImageInfo->imageView);
		}


		VK_CHECK_RESULT(coffeeSwapChain->QueuePresent(coffeeDevice->queue, 
		                                              currentBuffer, 
													  NULL));

		
		VK_CHECK_RESULT(vkQueueWaitIdle(coffeeDevice->queue));
	}

	void UpdateUniformBuffers()
	{
		ubo.model      = Matrix4::SetRotationAxis(-rotation.y, Vector3(0,1,0)) *
					     Matrix4::SetRotationAxis(-rotation.x, Vector3(1,0,0));

		ubo.projection = Matrix4::SetPerspectiveRH( 1.6f,//45, radians not degrees
											        (float)width / (float)height,
											        2.5f, 
											        400.0f);

		ubo.view       = Matrix4::SetLookAtRH( Vector3(0,4,-4), 
										       Vector3(0,0,0),
										       Vector3(0,1,0));

		uint8_t *pData;
		VK_CHECK_RESULT(vkMapMemory(coffeeDevice->device, uniformData.memory, 0, sizeof(ubo), 0, (void **)&pData));
		memcpy(pData, &ubo, sizeof(ubo));
		vkUnmapMemory(coffeeDevice->device, uniformData.memory);
	}





	HWND SetupWindow(HINSTANCE hinstance, WNDPROC wndproc)
	{
		this->windowInstance	= hinstance;

		WNDCLASSEX wndClass;
		wndClass.cbSize         = sizeof(WNDCLASSEX);
		wndClass.style          = CS_HREDRAW | CS_VREDRAW;
		wndClass.lpfnWndProc    = wndproc;
		wndClass.cbClsExtra     = 0;
		wndClass.cbWndExtra     = 0;
		wndClass.hInstance      = hinstance;
		wndClass.hIcon          = LoadIcon(NULL, IDI_APPLICATION);
		wndClass.hCursor        = LoadCursor(NULL, IDC_ARROW);
		wndClass.hbrBackground  = (HBRUSH)GetStockObject(BLACK_BRUSH);
		wndClass.lpszMenuName   = NULL;
		wndClass.lpszClassName  = name;
		wndClass.hIconSm        = LoadIcon(NULL, IDI_WINLOGO);

		if (!RegisterClassEx(&wndClass))
		{
			DBG_ASSERT_MSG(false, "Could not register window class!");
		}

		DWORD dwExStyle   = WS_EX_APPWINDOW | WS_EX_WINDOWEDGE;
		DWORD dwStyle     = WS_OVERLAPPEDWINDOW | WS_CLIPSIBLINGS | WS_CLIPCHILDREN;

		RECT windowRect;
		windowRect.left   = 0L;
		windowRect.top    = 0L;
		windowRect.right  = (long)width;
		windowRect.bottom = (long)height;

		AdjustWindowRectEx(&windowRect, dwStyle, FALSE, dwExStyle);

		std::string windowTitle = "Vulkan";
		windowHWND = CreateWindowEx(0,
			name,
			windowTitle.c_str(),
			dwStyle | WS_CLIPSIBLINGS | WS_CLIPCHILDREN,
			0,
			0,
			windowRect.right - windowRect.left,
			windowRect.bottom - windowRect.top,
			NULL,
			NULL,
			hinstance,
			NULL);


		// Center on screen
		uint32_t x = (GetSystemMetrics(SM_CXSCREEN) - windowRect.right) / 2;
		uint32_t y = (GetSystemMetrics(SM_CYSCREEN) - windowRect.bottom) / 2;
		SetWindowPos(windowHWND, 0, x, y, 0, 0, SWP_NOZORDER | SWP_NOSIZE);

		DBG_ASSERT_MSG(windowHWND, "Could not create window!");

		ShowWindow(windowHWND, SW_SHOW);
		SetForegroundWindow(windowHWND);
		SetFocus(windowHWND);

		return windowHWND;
	}


	void handleMessages(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
	{
		switch (uMsg)
		{
			case WM_CLOSE: 
			{
				// posts a WM_QUIT message
				PostQuitMessage( 0 );
			}
			break;

			case WM_PAINT:
				ValidateRect(windowHWND, NULL);
			break;

			case WM_KEYDOWN:
			break;

			case WM_RBUTTONDOWN:
			case WM_LBUTTONDOWN:
			case WM_MBUTTONDOWN:
			mousePos.x = (float)LOWORD(lParam);
			mousePos.y = (float)HIWORD(lParam);
			break;

			case WM_MOUSEWHEEL:
			{
				short wheelDelta = GET_WHEEL_DELTA_WPARAM(wParam);
				zoom += (float)wheelDelta * 0.005f * zoomSpeed;
			}
			break;

			case WM_MOUSEMOVE:
			if (wParam & MK_RBUTTON)
			{
				int32_t posx = LOWORD(lParam);
				int32_t posy = HIWORD(lParam);
				zoom += (mousePos.y - (float)posy) * .005f * zoomSpeed;
				mousePos = Vector3((float)posx, (float)posy, 0);
			}
			if (wParam & MK_LBUTTON)
			{
				int32_t posx = LOWORD(lParam);
				int32_t posy = HIWORD(lParam);
				rotation.x += (mousePos.y - (float)posy) * 1.25f * rotationSpeed;
				rotation.y -= (mousePos.x - (float)posx) * 1.25f * rotationSpeed;
				mousePos = Vector3((float)posx, (float)posy, 0);
			}
			if (wParam & MK_MBUTTON)
			{
				int32_t posx = LOWORD(lParam);
				int32_t posy = HIWORD(lParam);
				mousePos.x = (float)posx;
				mousePos.y = (float)posy;
			}
			break;
		}
	}	


	void StartRenderLoop()
	{
		MSG msg;
		while (TRUE)
		{
			while (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
			{
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}

			if (msg.message == WM_QUIT)
			{
				break;
			}

			Render();
		}

		// Flush coffeeDevice->device to make sure all resources can be freed 
		vkDeviceWaitIdle(coffeeDevice->device);
	}
};




static
VulkanDemo *vulkanDemo = NULL;				
														
LRESULT CALLBACK WndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)						
{																									
	if (vulkanDemo != NULL)																		
	{																								
		vulkanDemo->handleMessages(hWnd, uMsg, wParam, lParam);									
	}																								
	return (DefWindowProc(hWnd, uMsg, wParam, lParam));												
}										

															
int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR pCmdLine, int nCmdShow)	
{																									
	vulkanDemo = new VulkanDemo();															
	vulkanDemo->SetupWindow(hInstance, WndProc);																														
	vulkanDemo->SetupVulkan();																		
	vulkanDemo->StartRenderLoop();																	
	delete(vulkanDemo);																			
	return 0;																						
}			